import java.util.PriorityQueue;
import java.util.Set;
import java.util.HashSet;
import java.util.List;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;

public class AStar {
   private AStar() {} //don't let anyone actually create us

   public static <T extends Job> Solution<T> search(Set<T> nodes, Heuristic h) {
      Solution<T> order = null;
      int nodesExpanded = 0;
      Set<Solution> visited = new HashSet<Solution>();
      ArrayList<Solution<T>> queue = new ArrayList<Solution<T>>();

      for (T n : nodes) { //add initial partial solution
         List<T> partialOrder = new ArrayList<T>();
         partialOrder.add(n);
         queue.add(new Solution<T>(partialOrder, h));
      }

      Solution<T> next;
      while (queue.size() > 0 && (next = queue.remove(0)) != null  &&
             (order == null || next.getCost() < order.getCost())) {
         nodesExpanded++;
         if (next.getCompleted().size() == nodes.size()) {
            System.out.println("potential solution = " + next);
            if (order == null || next.getCost() < order.getCost()) {
               order = next;
            }
         } else if (!visited.contains(next)) {
            List<T> completed = next.getOrder();
            Set<T> remaining = difference(nodes, completed);
            for (T n : remaining) {
               //create a new child search node
               List<T> partialOrder = new ArrayList<T>(completed);
               partialOrder.add(n);
               Solution<T> child = new Solution<T>(partialOrder, h);

               //the verdict is out on this child
               if (!visited.contains(child)) {

                  //there is an equivalent solution already pending
                  if (queue.contains(child)) {
                     //so grab the original
                     Solution<T> competitor = queue.remove(queue.indexOf(child));

                     //and fight to the death
                     //System.out.print("orig " + competitor + " vs new " + child + "? ");
                     if (competitor.getCost() < child.getCost()) { //we are worse
                        //System.out.println("keep original");
                        child = competitor; //so put original back in
                     } else {
                        //System.out.println("replace");
                     }
                  } else {
                     //System.out.println("new: " + child);
                  }

                  //and let the favourite child win
                  queue.add(child);
               }
            }
            Collections.sort(queue);
         }
      }

      if (order != null) {
         order.setNodesExpanded(nodesExpanded);
      }
      return order;
   }

   private static <T> Set<T> difference(Set<T> all, Collection<T> minus) {
      Set<T> copy = new HashSet<T>(all);
      copy.removeAll(minus);
      return copy;
   }
}

