import java.util.Set;
import java.util.HashSet;
import java.awt.Point;
import java.util.Scanner;
import java.io.FileReader;
import java.io.FileNotFoundException;
import java.util.NoSuchElementException;
import java.util.InputMismatchException;
import java.lang.IllegalStateException;

public class CourierDelivery {

   private CourierDelivery() {} //not allowed to create this class

   public static boolean solve(Scanner s) throws Exception {
      Set<Job> jobs = getJobs(s);
      Solution order = AStar.search(jobs, new CourierHeuristic(jobs));

      boolean success = true;
      if (order != null) {
         order.print(System.out);

      } else { //no solution :(
         success = false;
      }

      return success;
   }

   private static Set<Job> getJobs(Scanner s) throws Exception {
      Set<Job> jobs = new HashSet<Job>();

      //more to do?
      while (s.hasNext()) {
         //skip the line prefix
         s.next("Job");

         //get the job start point
         Point start = new Point(s.nextInt(), s.nextInt());

         //skip the text in the middle
         s.next("to");

         //get the job end point
         Point end   = new Point(s.nextInt(), s.nextInt());

         //create the new job and add it to our list (set)
         jobs.add(new Job(start, end));

         //skip to the end of the line
         s.nextLine();
      }
      return jobs;
   }

   public static void main(String[] args) throws Exception {
      if (args.length > 0) {
         try {
            Scanner sc = new Scanner(new FileReader(args[0]));
            if (!CourierDelivery.solve(sc)) {
               System.out.println("no solution :(");
            }

         } catch (FileNotFoundException e) {
            System.err.println("File '" + args[0] + "' not found");
         } catch (InputMismatchException e) {
            System.err.println("Not a valid job file (unexpected token) "+e.getMessage());
         } catch (NoSuchElementException e) {
            System.err.println("Not a valid job file (premature end) " + e.getMessage());
         //} catch (Exception e) {
         //   System.err.println("Anonymous failure: " + e.getMessage());
         }
      } else {
         System.err.println("usage: javac CourierDelivery job_file");
      }
   }

}

