import java.util.Set;
import java.util.HashSet;
import java.util.List;
import java.awt.Point;
import java.io.PrintStream;

public class Solution<T extends Job> implements Comparable<Solution<T>> {

   public Solution(List<T> completed, Heuristic heuristic) {
      assert(completed != null);
      assert(heuristic != null);

      this.order = completed;
      this.completed = new HashSet<T>(completed);

      this.heuristic = heuristic;
      updateCost();
   }

   public void print(PrintStream out) {
      out.println(nodesExpanded + " nodes explored");
      out.println("cost = " + cost);

      Point current = new Point(0, 0); //starting place
      for (Job j : order) {
         Point start = j.start();
         Point finish = j.end();

         if (!start.equals(current)) { //don't "move" if we don't have to
            out.println("Move from " + current.x + " " + current.y +
                        " to " + start.x + " " + start.y);
         }

         out.println("Carry from " + start.x + " " + start.y +
                     " to " + finish.x + " " + finish.y);

         current = finish;
      }
   }

   public int compareTo(Solution<T> other) {
      return new Integer(getCost()).compareTo(other.getCost());
   }

   public boolean equals(Object other) {
      try {
         Solution<?> o = (Solution<?>) other;
         return equivalent(o);

      } catch (java.lang.ClassCastException e) {}
      return false;
   }

   public boolean equivalent(Solution<?> other) {
      int last = order.size() - 1;
      return completed.equals(other.completed) && //make sure we have the same nodes
             order.get(last).end().equals(other.order.get(last).end()); //and the same end point
   }

   public int hashCode() {
      return completed.hashCode();
   }

   public void setNodesExpanded(int num) {
      nodesExpanded = num;
   }

   public int getCost() {
      return cost + guess;
   }

   public String toString() {
      String s = "{" + getCost() + " = " + cost + " + " + guess + " | ";
      for (Job j : order) {
         s += "[" + j.start().x + " " + j.start().y + " - " + j.end().x + " " + j.end().y + "]";
      }
      s += "}";
      return s;
   }

   private void updateCost() {
      Point current = new Point(0, 0);
      cost = 0;

      for (Job j : order) {
         cost += heuristic.distance(current, j.start());
         cost += heuristic.distance(j.start(), j.end());

         current = j.end();
      }
      guess = heuristic.estimate(completed, order.get(order.size() - 1).end());
   }

   public Set<T> getCompleted() {
      return completed;
   }

   public List<T> getOrder() {
      return order;
   }

   private int nodesExpanded = 0; //hack! :(
   private Heuristic heuristic;
   private int cost = 0;
   private int guess = 0;
   private Set<T> completed;
   private List<T> order;

}

